/*******************************************************************************
 * Copyright (c) 2021, Bayerische Motoren Werke Aktiengesellschaft (BMW AG)
 *
 * This program and the accompanying materials are made
 * available under the terms of the Eclipse Public License 2.0
 * which is available at https://www.eclipse.org/legal/epl-2.0/
 *
 * SPDX-License-Identifier: EPL-2.0
 *******************************************************************************/

#include "TeleportAction.h"

#include <iostream>

namespace OPENSCENARIO
{

TeleportAction::TeleportAction(std::shared_ptr<NET_ASAM_OPENSCENARIO::v1_1::ITeleportAction> teleport_action_data,
                               std::shared_ptr<mantle_api::IEnvironment> environment,
                               const std::vector<std::string>& actors)
    : PrivateAction(environment, actors), teleport_action_data_(teleport_action_data)
{
}

void TeleportAction::Step()
{
    if (isRunning())
    {
        if (!teleport_action_data_->GetPosition())
        {
            throw std::runtime_error("TeleportAction does not contain Position");
        }

        std::shared_ptr<NET_ASAM_OPENSCENARIO::v1_1::IPosition> pos = teleport_action_data_->GetPosition();
        mantle_api::Pose pose = ConvertScenarioPosition(pos);

        for (auto actor : actors_)
        {
            environment_->GetEntityRepository().Get(actor).value().get().SetPosition(pose.position);
            environment_->GetEntityRepository().Get(actor).value().get().SetOrientation(pose.orientation);
        }
        End();
    }
}

// TODO: Move this to a common class when needed for other actions/conditions as well
mantle_api::Pose TeleportAction::ConvertScenarioPosition(std::shared_ptr<NET_ASAM_OPENSCENARIO::v1_1::IPosition> pos)
{
    mantle_api::Position position;
    mantle_api::Pose pose;
    if (auto scenario_lane_position = pos->GetLanePosition())
    {
        mantle_api::OpenDrivePosition lane_position;
        lane_position.referenced_object.road = atoi(scenario_lane_position->GetRoadId().c_str());
        lane_position.referenced_object.lane = atoi(scenario_lane_position->GetLaneId().c_str());
        lane_position.s_offset = units::length::meter_t(scenario_lane_position->GetS());
        lane_position.t_offset = units::length::meter_t(scenario_lane_position->GetOffset());
        position = lane_position;
    }
    else if (auto geo_position = pos->GetGeoPosition())
    {
        position = mantle_api::LatLonPosition{units::angle::radian_t(geo_position->GetLatitude()),
                                              units::angle::radian_t(geo_position->GetLongitude())};

        auto orientation = geo_position->GetOrientation();
        if (orientation != nullptr)
        {
            pose.orientation =
                mantle_api::Orientation3<units::angle::radian_t>{units::angle::radian_t(orientation->GetH()),
                                                                 units::angle::radian_t(orientation->GetP()),
                                                                 units::angle::radian_t(orientation->GetR())};
        }
    }
    else
    {
        std::cout << "Conversion from this type of OpenSCENARIO position to mantle API position not implemented yet"
                  << std::endl;
    }

    const auto* mantle_converter = environment_->GetConverter();
    pose.position = mantle_converter->Convert(position);
    return pose;
}

}  // namespace OPENSCENARIO
